CST_Intbc <- function(exp, obs, target_grid, int_method, bc_method, points = NULL, method_point_interp = NULL, 
                      lat_dim = "lat", lon_dim = "lon", sdate_dim = "sdate", member_dim = "member", 
                      remap_region = NULL, cal_method = "mse_min", ncores = 1)
{
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }

  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  res <- Intbc(exp = exp$data, obs = obs$data, exp_lats = exp[[lat_dim]], exp_lons = exp[[lon_dim]], 
               obs_lats = obs[[lat_dim]], obs_lons = obs[[lon_dim]], target_grid = target_grid, 
               int_method = int_method, bc_method = bc_method, points = points, 
               method_point_interp = method_point_interp, lat_dim = lat_dim, lon_dim = lon_dim, 
               sdate_dim = sdate_dim, member_dim = member_dim, remap_region = remap_region, 
               cal_method = cal_method, ncores = ncores)

  res_s2dv <- suppressWarnings(s2dv_cube(data = res$data, lon = res$lon, lat = res$lat)) 

  return(res_s2dv)
}


#'@rdname Intbc
#'@title Downscaling using interpolation and bias adjustment.
#' 
#'@author J. Ramon, \email{jaume.ramon@bsc.es}
#'
#'@description This function performs a downscaling using an interpolation and a later bias 
#'adjustment. It is recommended that the observations are passed already in the target grid. 
#'Otherwise, the function will also perform an interpolation of the observed field into the 
#'target grid. The coarse scale and observation data can be either global or regional. In the 
#'latter case, the region is defined by the user. In principle, the coarse and observation data 
#'are intended to be of the same variable, although different variables can also be admitted. 
#'
#'@param exp an 's2dv_cube' object containing the experimental field on the coarse scale for 
#'which the downscaling is aimed. The element 'data' in the 's2dv_cube' object must have, at 
#'least, the dimensions latitude, longitude, start date and member. The object is expected to 
#'be already subset for the desired large scale region. Latitudes must range from -90 to 90 
#'(or a subset) and longitudes must range from -180 to 180.
#'@param obs an 's2dv_cube' object containing the observational field. The element 'data' in 
#'the 's2dv_cube' object must have, at least, the dimensions latitude, longitude and start date. 
#'The object is expect to be already subset for the desired large scale region.
#'@param int_method a character vector indicating the regridding method to be passed to CDORemap. 
#'Accepted methods are "con", "bil", "bic", "nn", "con2". If "nn" method is to be used, CDO_1.9.8 
#'or newer version is required. 
#'@param bc_method a character vector indicating the bias adjustment method to be applied after
#'the interpolation. Accepted methods are 'simple_bias', 'calibration', 'quantile_mapping'. The 
#'abbreviations 'sbc', 'cal', 'qm' can also be used.
#'@param points a list of two elements containing the point latitudes and longitudes of the 
#'locations to downscale the model data. The list must contain the two elements 'lat' and 'lon'. 
#'If the downscaling is to a point location, only regular grids are allowed for exp and obs. 
#'@param method_point_interp a character vector indicating the interpolation method to interpolate 
#'model gridded data into the point locations. Accepted methods are "nearest", "bilinear", "9point", 
#'"invdist4nn", "NE", "NW", "SE", "SW".
#'@param lat_dim a character vector indicating the latitude dimension name in the element 'data' 
#'in exp and obs. Default set to "lat".
#'@param lon_dim a character vector indicating the longitude dimension name in the element 'data' 
#'in exp and obs. Default set to "lon".
#'@param sdate_dim a character vector indicating the start date dimension name in the element 
#''data' in exp and obs. Default set to "sdate".
#'@param member_dim a character vector indicating the member dimension name in the element
#''data' in exp and obs. Default set to "member". 
#'@param ncores an integer indicating the number of cores to use in parallel computation.
#'
#'@import CSTools 
#'
#'@seealso \code{\link[CSTools]{BiasCorrection}}
#'@seealso \code{\link[CSTools]{Calibration}}
#'@seealso \code{\link[CSTools]{QuantileMapping}}
#'
#'@examples 
#'exp <- rnorm(500) 
#'dim(exp) <- c(member = 5, lat = 4, lon = 5, sdate = 5) 
#'lon_exp <- 1:5 
#'lat_exp <- 1:4 
#'obs <- rnorm(900) 
#'dim(obs) <- c(lat = 12, lon = 15, sdate = 5) 
#'lon_obs <- seq(1,5, 4/14) 
#'lat_obs <- seq(1,4, 3/11) 
#'exp <- s2dv_cube(data = exp, lat = lat_exp, lon = lon_exp) 
#'obs <- s2dv_cube(data = obs, lat = lat_obs, lon = lon_obs)
#'attr(exp$lon,"first_lon") <- 1 
#'attr(exp$lon,"last_lon") <- 5 
#'attr(exp$lat,"first_lat") <- 1 
#'attr(exp$lat,"last_lat") <- 4 
#'attr(obs$lon,"first_lon") <- 1 
#'attr(obs$lon,"last_lon") <- 5 
#'attr(obs$lat,"first_lat") <- 1 
#'attr(obs$lat,"last_lat") <- 4
#'downscaled_field <- Intbc(exp = exp, obs = obs, target_grid = 'r1280x640', bc_method = 'simple_bias', int_method = 'conservative')
#'@export
Intbc <- function(exp, obs, exp_lats, exp_lons, obs_lats, obs_lons, target_grid, int_method, bc_method, 
                  points = NULL, method_point_interp = NULL, lat_dim = "lat", lon_dim = "lon", sdate_dim = "sdate", 
                  member_dim = "member", remap_region = NULL, cal_method = "mse_min", ncores = 1) {

  if (!inherits(int_method, 'character')) {
    stop("Parameter 'int_method' must be of the class 'character'")
  }
 
  if (!inherits(bc_method, 'character')) {
    stop("Parameter 'bc_method' must be of the class 'character'")
  }

  if (!inherits(lat_dim, 'character')) {
    stop("Parameter 'lat_dim' must be of the class 'character'")
  }

  if (!inherits(lon_dim, 'character')) {
    stop("Parameter 'lon_dim' must be of the class 'character'")
  }

  if (!inherits(sdate_dim, 'character')) {
    stop("Parameter 'sdate_dim' must be of the class 'character'")
  }

  if (!inherits(member_dim, 'character')) {
    stop("Parameter 'member_dim' must be of the class 'character'")
  }

  # Do not allow synonims for lat (latitude), lon (longitude) and time (sdate) dimension names 
  if (is.na(match(lon_dim, names(dim(exp)))) | is.na(match(lon_dim, names(dim(obs))))) {
    stop("Missing longitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lon_dim'")
  }

  if (is.na(match(lat_dim, names(dim(exp)))) | is.na(match(lat_dim, names(dim(obs))))) {
    stop("Missing latitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lat_dim'")
  }

  if (is.na(match(sdate_dim, names(dim(exp)))) | is.na(match(sdate_dim, names(dim(obs))))) {
    stop("Missing start date dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'sdate_dim'")
  }

  if (is.na(match(member_dim, names(dim(exp))))) {
    stop("Missing member dimension in 'exp', or does not match the parameter 'member_dim'")
  }

  if (!(bc_method %in% c('sbc', 'cal', 'qm', 'simple_bias', 'calibration', 'quantile_mapping'))) {
    stop("Parameter 'bc_method' must be a character vector indicating the bias adjustment method. ",
         "Accepted methods are 'simple_bias', 'calibration', 'quantile_mapping'. The abbreviations ",
         "'sbc', 'cal', 'qm' can also be used.")
  }

  if (is.null(remap_region)) {
    remap_region <- c(obs_lons[1], obs_lons[length(obs_lons)], obs_lats[1], obs_lats[length(obs_lats)])
  }

  exp_interpolated <- Interpolation(exp = exp, lats = exp_lats, lons = exp_lons, target_grid = target_grid, 
                                    method_remap = int_method, points = points, 
                                    method_point_interp = method_point_interp, remap_region = remap_region)

  # If after interpolating 'exp' data the coordinates do not match, the obs data is interpolated to
  # the same grid to force the matching
  if (!.check_coords(lat1 = exp_lats, lat2 = obs_lats, lon1 = exp_lons, lon2 = obs_lons) | !is.null(points)) {
    obs_interpolated <- Interpolation(exp = obs, lats = obs_lats, lons = obs_lons, target_grid = target_grid, 
                                      method_remap = int_method, points = points, 
                                      method_point_interp = method_point_interp, remap_region = remap_region)
  } else {
    obs_interpolated <- obs
  }

  #.check_coords
  if (bc_method == 'sbc' | bc_method == 'simple_bias') {
    if (dim(obs_interpolated$data)[sdate_dim] == 1) {
      warning('Simple Bias Correction should not be used with 
              only one observation. Returning NA. ')
    }
    res <- BiasCorrection(exp = exp_interpolated$data, 
                          obs = obs_interpolated$data) 
  } 
  else if (bc_method == 'cal' | bc_method == 'calibration') {
    if (dim(exp_interpolated$data)[member_dim] == 1) {
      stop('Calibration must not be used with
           only one ensemble member.')
    }
    res <- Calibration(exp = exp_interpolated$data, obs = obs_interpolated$data, 
                       cal.method = cal_method, ncores = ncores)
  }
  else if (bc_method == 'qm' | bc_method == 'quantile_mapping') {
    if (any(is.na(exp_interpolated$data))) {
      warning('Found NAs in "exp" data, either introduced with the initial object or ', 
              'by the interpolation method. Quantile Mapping method is likely to fail.')
    }
    res <- QuantileMapping(exp = exp_interpolated$data, obs = obs_interpolated$data, 
                           sample_dims = sdate_dim, method = 'QUANT', ncores = ncores)
  }

  # Create an s2dv_cube object
  res <- list(data = res, lon = exp_interpolated$lon, lat = exp_interpolated$lat)

  return(res)
}




